@@ -23,6 +23,7 @@ module Hledger.Utils.IO (
   error',
   usageError,
   warn,
+  warnIO,
   ansiFormatError,
   ansiFormatWarning,
   printError,
@@ -45,13 +46,13 @@ module Hledger.Utils.IO (
   readFileOrStdinPortably',
   readFileStrictly,
   readFilePortably,
-  readHandlePortably,
-  readHandlePortably',
+  hGetContentsPortably,
   -- hereFileRelative,
-  inputToHandle,
+  textToHandle,
 
   -- * Command line parsing
   progArgs,
+  getFlag,
   getOpt,
   parseYN,
   parseYNA,
@@ -130,12 +131,11 @@ import           Data.FileEmbed (makeRelativeToProject, embedStringFile)
 import           Data.Functor ((<&>))
 import           Data.List hiding (uncons)
 import           Data.Maybe (isJust, catMaybes)
-import           Data.Ord (comparing, Down (Down))
-import qualified Data.Text as T
+import Data.Text qualified as T
 import           Data.Text.Encoding.Error (UnicodeException)
-import qualified Data.Text.IO as T
-import qualified Data.Text.Lazy as TL
-import qualified Data.Text.Lazy.Builder as TB
+import Data.Text.IO qualified as T
+import Data.Text.Lazy qualified as TL
+import Data.Text.Lazy.Builder qualified as TB
 import           Data.Time.Clock (getCurrentTime)
 import           Data.Time.LocalTime (LocalTime, ZonedTime, getCurrentTimeZone, utcToLocalTime, utcToZonedTime)
 import           Data.Word (Word16)
@@ -153,13 +153,14 @@ import           System.Exit (exitFailure)
 import           System.FilePath (isRelative, (</>))
 import "Glob"    System.FilePath.Glob (glob)
 import           System.Info (os)
-import           System.IO (Handle, IOMode (..), hClose, hGetEncoding, hIsTerminalDevice, hPutStr, hPutStrLn, hSetNewlineMode, hSetEncoding, openFile, stderr, stdin, stdout, universalNewlineMode, utf8_bom)
-import qualified System.IO.Encoding as Enc
+import           System.IO (Handle, IOMode (..), hClose, hGetEncoding, hIsTerminalDevice, hPutStr, hPutStrLn, hSetNewlineMode, hSetEncoding, openFile, stderr, stdin, stdout, universalNewlineMode, utf8_bom, utf8)
+import System.IO.Encoding qualified as Enc
 import           System.IO.Unsafe (unsafePerformIO)
 import           System.Process (CreateProcess(..), StdStream(CreatePipe), createPipe, shell, waitForProcess, withCreateProcess)
 import           Text.Pretty.Simple (CheckColorTty(..), OutputOptions(..), defaultOutputOptionsDarkBg, defaultOutputOptionsNoColor, pShowOpt, pPrintOpt)
 
 import Hledger.Utils.Text (WideBuilder(WideBuilder))
+import Control.Monad.IO.Class (MonadIO, liftIO)
 
 
 -- Pretty showing/printing
@@ -217,16 +218,20 @@ usageError = error' . (++ " (use -h to see usage)")
 ansiFormatError :: String -> String
 ansiFormatError = (<> sgrresetall) . ((sgrbrightred <> sgrbold) <>)
 
--- | Show a message, with "Warning:" label, on stderr before returning the given value.
--- Also do some ANSI styling of the first line when allowed (using unsafe IO).
--- Currently we use this very sparingly in hledger; we prefer to either quietly work,
--- or loudly raise an error. (Varying output can make scripting harder.)
+-- | Show a warning message on stderr before returning the given value.
+-- Like trace, but prepends a "Warning:" label, and does some ANSI styling of the first line when allowed (using unsafe IO).
+-- Currently we use this very sparingly in hledger; we prefer to either quietly work, or loudly raise an error.
+-- Varying output can make scripting harder. But on stderr, it shouldn't cause much hassle.
 warn :: String -> a -> a
-warn msg = trace msg'
-  where
-    msg' =
-      (if useColorOnStderrUnsafe then modifyFirstLine ansiFormatWarning else id) $
-      "Warning: "<> msg
+warn = trace . formatWarning
+
+-- | Like warn, but take extra care to sequence properly in IO.
+warnIO :: MonadIO m => String -> m ()
+warnIO = liftIO . traceIO . formatWarning
+
+formatWarning =
+  (if useColorOnStderrUnsafe then modifyFirstLine ansiFormatWarning else id) .
+  ("Warning: " <>)
 
 -- | Apply standard ANSI SGR formatting (yellow, bold) suitable for console warning text.
 ansiFormatWarning :: String -> String
@@ -238,9 +243,8 @@ ansiFormatWarning = (<> sgrresetall) . ((sgrbrightyellow <> sgrbold) <>)
 modifyFirstLine :: (String -> String) -> String -> String
 modifyFirstLine f s = intercalate "\n" $ map f l <> ls where (l,ls) = splitAt 1 $ lines s  -- total
 
-{- | Print an error message to stderr, with a consistent "programname: " prefix,
-and applying ANSI styling (bold bright red) to the first line if that is supported and allowed.
--}
+-- | Print an error message to stderr, with a consistent "programname: " prefix,
+-- and applying ANSI styling (bold bright red) to the first line if that is supported and allowed.
 printError :: String -> IO ()
 printError msg = do
   progname <- getProgName
@@ -256,9 +260,8 @@ printError msg = do
         <> (if "Error:" `isPrefixOf` msg then "" else "Error: ")
   hPutStrLn stderr $ style $ prefix <> msg
 
-{- | Print an error message with printError,
-then exit the program with a non-zero exit code.
--}
+-- | Print an error message with printError,
+-- then exit the program with a non-zero exit code.
 exitWithErrorMessage :: String -> IO ()
 exitWithErrorMessage msg = printError msg >> exitFailure
 
@@ -337,7 +340,8 @@ handleExit = flip catches [
         noencoding = map toLower enc == "ascii"
         msg = unlines $ [
             rstrip $ show ex
-          , "Some text could not be decoded with the system text encoding, " <> enc <> "."
+          , "Some text could not be decoded with the system's text encoding, " <> enc
+          , "(or, the text encoding specified by CSV rules)."
           ] ++
           if noencoding
           then [
@@ -402,7 +406,7 @@ expandHomePath = \case
     ('~':_)      -> ioError $ userError "~USERNAME in paths is not supported"
     p            -> return p
 
--- | Given a current directory, convert a possibly relative, possibly tilde-containing
+-- | Given a current directory, convert a possibly relative, possibly tilde-prefixed
 -- file path to an absolute one.
 -- ~username is not supported. Leaves "-" unchanged. Can raise an error.
 expandPath :: FilePath -> FilePath -> IO FilePath -- general type sig for use in reader parsers
@@ -412,14 +416,15 @@ expandPath curdir p = (if isRelative p then (curdir </>) else id) <$> expandHome
 -- | Like expandPath, but treats the expanded path as a glob, and returns
 -- zero or more matched absolute file paths, alphabetically sorted.
 -- Can raise an error.
+-- For a more elaborate glob expander, see 'findMatchedFiles' (used by the include directive).
 expandGlob :: FilePath -> FilePath -> IO [FilePath]
 expandGlob curdir p = expandPath curdir p >>= glob <&> sort  -- PARTIAL:
 
--- | Given a list of existing file paths, sort them by modification time, most recent first.
+-- | Given a list of existing file paths, sort them by modification time (from oldest to newest).
 sortByModTime :: [FilePath] -> IO [FilePath]
 sortByModTime fs = do
   ftimes <- forM fs $ \f -> do {t <- getModificationTime f; return (t,f)}
-  return $ map snd $ sortBy (comparing Data.Ord.Down) ftimes
+  return $ map snd $ sort ftimes
 
 -- | Like readFilePortably, but read all of the file before proceeding.
 readFileStrictly :: FilePath -> IO T.Text
@@ -430,7 +435,7 @@ readFileStrictly f = readFilePortably f >>= \t -> evaluate (T.length t) >> retur
 -- using the system locale's text encoding,
 -- ignoring any utf8 BOM prefix (as seen in paypal's 2018 CSV, eg) if that encoding is utf8.
 readFilePortably :: FilePath -> IO T.Text
-readFilePortably f =  openFile f ReadMode >>= readHandlePortably
+readFilePortably f = openFile f ReadMode >>= hGetContentsPortably Nothing
 
 -- | Like readFilePortably, but read from standard input if the path is "-".
 readFileOrStdinPortably :: String -> IO T.Text
@@ -438,7 +443,7 @@ readFileOrStdinPortably = readFileOrStdinPortably' Nothing
 
 -- | Like readFileOrStdinPortably, but take an optional converter.
 readFileOrStdinPortably' :: Maybe DynEncoding -> String -> IO T.Text
-readFileOrStdinPortably' c f = openFileOrStdin f >>= readHandlePortably' c
+readFileOrStdinPortably' c f = openFileOrStdin f >>= hGetContentsPortably c
 
 -- | Open a file for reading, using the standard System.IO.openFile.
 -- This opens the handle in text mode, using the initial system locale's text encoding.
@@ -446,32 +451,27 @@ openFileOrStdin :: String -> IO Handle
 openFileOrStdin "-" = return stdin
 openFileOrStdin f' = openFile f' ReadMode
 
--- readHandlePortably' with no text encoding specified.
-readHandlePortably :: Handle -> IO T.Text
-readHandlePortably = readHandlePortably' Nothing
-
--- | Read text from a handle with a specified encoding, using the encoding package.
--- Or if no encoding is specified, it uses the handle's current encoding,
--- after first changing it to UTF-8BOM if it was UTF-8, to allow a Byte Order Mark at the start.
+-- | Read text from a handle, perhaps using a specified encoding from the encoding package.
+-- Or if no encoding is specified, using the handle's current encoding,
+-- changing it to UTF-8BOM if it was UTF-8, to ignore any Byte Order Mark at the start.
 -- Also it converts Windows line endings to newlines.
 -- If decoding fails, this throws an IOException (or possibly a UnicodeException or something else from the encoding package).
-readHandlePortably' :: Maybe DynEncoding -> Handle -> IO T.Text
-readHandlePortably' Nothing h = do
+hGetContentsPortably :: Maybe DynEncoding -> Handle -> IO T.Text
+hGetContentsPortably Nothing h = do
   hSetNewlineMode h universalNewlineMode
   menc <- hGetEncoding h
   when (fmap show menc == Just "UTF-8") $ hSetEncoding h utf8_bom
   T.hGetContents h
-readHandlePortably' (Just e) h =
+hGetContentsPortably (Just e) h =
   -- convert newlines manually, because Enc.hGetContents uses bytestring's hGetContents
   T.replace "\r\n" "\n" . T.pack <$> let ?enc = e in Enc.hGetContents h
 
--- | Create a handle from which the given text can be read.
--- Its encoding will be UTF-8BOM.
-inputToHandle :: T.Text -> IO Handle
-inputToHandle t = do
+-- | Create a handle from which the given text can be read. Its encoding will be UTF-8.
+textToHandle :: T.Text -> IO Handle
+textToHandle t = do
   (r, w) <- createPipe
-  hSetEncoding r utf8_bom
-  hSetEncoding w utf8_bom
+  hSetEncoding r utf8
+  hSetEncoding w utf8
   -- use a separate thread so that we don't deadlock if we can't write all of the text at once
   forkIO $ T.hPutStr w t >> hClose w
   return r
@@ -502,6 +502,15 @@ progArgs = unsafePerformIO getArgs
 --  a few cases involving --color (see useColorOnStdoutUnsafe)
 --  --debug
 
+-- | Given one or more long or short flag names,
+-- report whether this flag is present in the command line.
+-- Concatenated short flags (-a -b written as -ab) are not supported.
+getFlag :: [String] -> IO Bool
+getFlag names = do
+  let flags = map toFlag names
+  args <- getArgs
+  return $ any (`elem` args) flags
+
 -- | Given one or more long or short option names, read the rightmost value of this option from the command line arguments.
 -- If the value is missing raise an error.
 -- Concatenated short flags (-a -b written as -ab) are not supported.
@@ -614,7 +623,6 @@ getTerminalWidth  = fmap snd <$> getTerminalHeightWidth
 --   --chop-long-lines
 --   --hilite-unread
 --   --ignore-case
---   --mouse
 --   --no-init
 --   --quit-at-eof
 --   --quit-if-one-screen
@@ -636,7 +644,6 @@ setupPager = do
        "--chop-long-lines"
       ,"--hilite-unread"
       ,"--ignore-case"
-      ,"--mouse"
       ,"--no-init"
       ,"--quit-at-eof"
       ,"--quit-if-one-screen"
@@ -765,7 +772,7 @@ useColorOnStderrUnsafe = unsafePerformIO useColorOnStderr
 
 -- | Detect whether ANSI should be used on stdout using useColorOnStdoutUnsafe,
 -- and if so prepend and append the given SGR codes to a string.
--- Currently used in a few places (the commands list, the demo command, the recentassertions error message);
+-- Currently used in a few places (the commands list, the recentassertions error message, add, demo);
 -- see useColorOnStdoutUnsafe's limitations.
 ansiWrapUnsafe :: SGRString -> SGRString -> String -> String
 ansiWrapUnsafe pre post s = if useColorOnStdoutUnsafe then pre<>s<>post else s
